/* See COPYING for license details
 * (C)2000 Andrew Kozin AKA Stanson */
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>

#include "common.h"
#include "real_modem.h"

extern int debug;

/*
 * transmitter schematic
 *
 * +--------------+  +-----------+  +-------------+
 * | control data |  | user data |  | RR/NR frame |
 * +------+-------+  +-----+-----+  +------+------+
 *        |                |               |
 *        +<---------------+               |
 *        |                                |
 *        |  +-------------+               |
 *        +->| Frame queue |-+             |
 *           1-------------+ |-+           V
 *             2-------------+ |-+   +-------------+   +-----------+
 *               3-------------+ |-->| Data buffer |-->| Bitstream |
 *                 4-------------+   +-------------+   +-----------+
 *
 *   Frame queue updated each time we receive RR frame. Sent frames are
 * removed, and new data is appended, if any.
 *
 *   Data buffer updated each time, when the Bitstream is empty. New
 * data copied from Frame queue or set to RR/NR frame, if any.
 */

/* syncronization sequence 
 * [NMT Doc 450-3, Annex 21 v.3.0.0, 4. Frame Structure] */

int sync_bits[] = 
{
  1,0,1,0,1,0,1,0,1,0,1,0,1,0,1, /* bit syncronization */
  0,0,1,0,1,0,0,0,1              /* frame syncronization word */
};

/*
 *            DT frame structure
 *
 * number  0  1  2  3  4  5  6  7  8  9 10 11
 * type    l  d  d  d  d  d  d  d  d  c  c  c
 * len     7  7  7  7  7  7  7  7  7  7  7  2
 *
 *  where type is
 *   l - Label
 *   d - data
 *   c - checksum
 */

/*
 * get_checksum - calculates checksum from frame.
 */

unsigned short get_checksum(unsigned char *data)
{unsigned long  cs;
 int            i,len;
  for(i = 0, cs = 0; i < 12; i++)
  {
    cs |= ((*data++) << 1) & 0xFE;
    len = (i == 11) ? 2 : 7;
    while(len--) { cs <<= 1; if(cs & 0x01000000) cs ^= 0x01102100; }
  }
  return (unsigned short)(cs >> 8);
}

/*
 * set_checksum - sets checksum for frame
 */
void set_checksum(unsigned char *data)
{unsigned short cs;
  memset(&data[9], 0, 3);
  cs = get_checksum(data);
  data[9]  = (cs >> 9) & 0x7F;
  data[10] = (cs >> 2) & 0x7F;
  data[11] = (cs << 5) & 0x60;
  return;
}

/*
 * get_parity - calculates parity for byte
 */

unsigned char get_parity(unsigned char c)
{unsigned char p; 
 int           i;
  for(p = 0, i = 0; i < 7; i++, c >>= 1) if(c & 1) p ^= 0x60;
  return p;
}

int fill_queue(struct _transmitter *tx)
{unsigned char *data, *pd, b;
 int i, len;

  while(tx->queue_len < 4)
  {
    data = &tx->queue[tx->queue_len][0];
    if(tx->control_flag)
    {
      memcpy(data + 1, &tx->control[0], 8);
      tx->control_flag = 0; /* we already eat control frame */
      data[0] = POSITIVE_RESPONSE; /* CT frame */
    }
    else
    {
      if(!tx->tty_len) return 0; /* no data to transmit */
      /* Set DT frame
       * As the DT frame length is fixed and NUL characters are used
       * for filling the last unused locations of frame, those NUL 
       * characters which are intended for transmission cannot be
       * situated as the last characters of the frame. In cases where
       * this would happen, a new DT frame shall be started with
       * the NUL as the first character.
       * [NMT Doc 450-3, Annex 21 v.3.0.0, 7.8 Dealing with NUL Characters] */
      len = tx->tty_len;                  /* get length of data to transmit */
      if(len > tx->wif) len = tx->wif;    /* cut it to max words in frame */
      while((len > 1) && !tx->tty_buf[len - 1]) len--; /* strip NUL's */
      pd = data + (9 - tx->wif);          /* get pointer to destination */
      memcpy(pd, &tx->tty_buf[0], len);   /* copy data */
      memset(pd + len, 0, tx->wif - len); /* fill the unused locations */
      tx->tty_len -= len;                 /* shift input buffer */
      memmove(&tx->tty_buf[0], &tx->tty_buf[len], tx->tty_len);
      if(tx->wif == 7)
      {/* for 8 bit mode copy MSB's to first data location */
        for(b = 0, i = 2; i < 9; i++, pd++)
        { 
          b <<= 1;
          if(*pd & 0x80) b |= 1;
        }
        data[1] = b;
      }
      data[0] = NEGATIVE_RESPONSE; /* DT frame */
    }
    data[0] |= DT_PREFIX | tx->dir | (tx->n & 7); /* set label */
    set_checksum(data); /* calculate checksum */
    tx->n++; /* increment frame number */
    tx->queue_len++;
  }
  return 1;
}

int tx_bit(struct _transmitter *tx)
{int           bit, i;
 unsigned char label;

  /* Hold timer prevents us from sending DT frames, but we still acknowledge
   * reception of DT frames with RR/NR. So we decrement it, if it is active,
   * and continue */
  if(tx->hold) tx->hold--;

  /* Check if we are transmitting frame. If true - go and get next bit */
  if(tx->frm_len) goto next_bit;

  /* Frame sent, do we have more? */
  tx->bit_cnt   = 0; /* reset counters */
  tx->word_cnt  = 0;

  if(tx->bit_num)
  { /* OK, this was not silence */
    /* When a DT frame follows directly another frame no bit syncronization
     * is sent.
     * [NMT Doc 450-3, Annex 21 v.3.0.0, 5.5. Flow Diagram and Timing] */
    tx->bit_num   = 15;
    if(((tx->data[0] & 0x18) != DT_PREFIX))
    {
      if(!tx->frm_cnt)
      {/* resend RR/NR frame without bit sync */
         tx->frm_cnt++;
         tx->frm_len = 31;
         goto next_bit;
      }
      else
      {/* enougth */
        tx->frm_cnt = 0;
      }
    }
  }

  if(tx->dt_label) /* DT(N) received, we must send RR(N+1) */
  {
    label = tx->dt_label;
    tx->dt_label = 0; /* clear label */
    /* Acknowledgement frames RR/NR are transmitted 22.5+-2 ms after
     * reception of a DT frame. However, in case an RR/NR follows another
     * frame (DT/RR/NR) it is sent with no delay.
     * [NMT Doc 450-3, Annex 21 v.3.0.0, 5.5. Flow Diagram and Timing] */
    if(!tx->bit_num) tx->delay = 27; /* 0.0225 * BAUDRATE */
    /* An RR/NR frame is always sent with the bit syncronization.
     * [NMT Doc 450-3, Annex 21 v.3.0.0, 5.5. Flow Diagram and Timing] */
    tx->bit_num = 0;
    tx->frm_len = 46; /* frame length in bits */
    tx->frm_cnt = 0;
    /* create RR(N+1) frame */
    tx->data[0] = ((label + 1) & 7) | tx->dir | RR_PREFIX | NEGATIVE_RESPONSE;
    tx->data[1] = tx->data[0];
    tx->data[2] = get_parity(tx->data[0]); /* set parity bits */
#ifdef DEBUG
    printf("<RR(%u)\n", tx->data[0] & 7);
#endif
    goto next_bit;
  }

  if(tx->rr_label)
  {
    label = tx->rr_label;
    tx->rr_label = 0; /* clear label */
    if((label & 0x18) == RR_PREFIX)
    { /* RR(N) received, we can send DT(N) frame */
      tx->frm_cnt   = 0;
      tx->hold      = 0; /* stop and clear timer */
      /* remove sent frames from queue */
      label         = ((label & 7) - tx->ntx) & 7;
      tx->ntx       = (tx->ntx + label) & 7;
      tx->queue_len -= label;
      memmove(&tx->queue[0][0],&tx->queue[label][0],12 * tx->queue_len);
      /* clear the rest of queue */
      memset(&tx->queue[tx->queue_len][0],0,4 - tx->queue_len);
    }
    else
    { /* NR(N) received, wait for 5 seconds */
      tx->hold    = 5 * BAUDRATE; /* set hold to 5 seconds and keep silence */
      tx->bit_num = 0;
    }
  }

  if(tx->hold) return -1;

  fill_queue(tx);

  /* check queue for ready frames */
  if(!tx->queue_len) {tx->bit_num = 0; return -1;}  /* we have nothing to transmit. Keep silence */
  memcpy(&tx->data[0], &tx->queue[tx->frm_cnt][0], 12);
  tx->frm_len = 127 - tx->bit_num; /* frame length in bits */
  if(++tx->frm_cnt >= tx->queue_len) tx->frm_cnt = 0;

#ifdef DEBUG
  printf("<%cT(%u) ", tx->data[0] & 0x20 ? 'D' : 'C', tx->data[0] & 7 );
  for(i = 1; i < 9; i++) printf("%02X ", tx->data[i]);
  printf("\n");
#endif

  /* transmit next bit */
next_bit:
  /* decrement RR/NR delay timer and keep silence until delay ends */
  if(tx->delay) {tx->delay--; return -1;} 

  if(tx->bit_num <= 23)
  {/* we are transmitting syncronization sequence */
    bit = sync_bits[tx->bit_num];
  }
  else
  {/* word structure: 11ddddddd */
    if(!tx->bit_cnt)
    {/* load stream and set 2nd bit to 1 */
      tx->stream  = tx->data[tx->word_cnt++] | 0x80;
      tx->bit_cnt = 8;     /* reload bit counter */
      bit = 1;             /* 1st bit is always 1 */
    }
    else
    {
      bit = (tx->stream & 0x80) ? 1 : 0;
      tx->stream <<= 1;
      tx->bit_cnt--;
    }
  }
  tx->bit_num++;
  tx->frm_len--;
  return bit;
}

int tx_init(struct _transmitter *tx)
{int i;
 double t;
  memset(tx,0,sizeof(struct _transmitter));
  /* create transmitter sine tables */
  for(i = 0; i < BAUD_TIME; i++)
  {
    t = (double)i/(double)SAMPLE_RATE;
    tx->sin_lo[i] = (short)((double)SHRT_MAX * sin(2.0*M_PI*1200.0*t));
    tx->sin_hi[i] = (short)((double)SHRT_MAX * sin(2.0*M_PI*1800.0*t));
  }
  tx->wif = 8; /* initially select 7-bit mode */
  return 0;
}

int tx_encode(struct _transmitter *tx, short *buf, int size)
{int   i;
 short sample;

  for(i = 0; i < size; i++, buf++)
  {
    if(!tx->si)
    {
      if(!tx->current_bit) tx->phase ^= 1;
      tx->current_bit = tx_bit(tx);
    }
    if(tx->current_bit < 0)
    {/* just silence */
      *buf = 0;
    }
    else
    {/* modulation */
      sample = tx->current_bit ? tx->sin_lo[tx->si] : tx->sin_hi[tx->si];
      *buf = tx->phase ? -sample : sample;
    }
    if(++tx->si >= BAUD_TIME) tx->si = 0;
  }
  return 0;
}
